#!/usr/bin/env python
# -*- coding: utf-8 -*-

# # Reconstruction de synonymes - correction
# 
# On s'intéresse la construction automatiquement de synonymes. En résumé :
# 
# * les synonymes via la base [WOLF](http://alpage.inria.fr/~sagot/wolf-en.html)
# * on ne garde que les synonymes composé d'un seul mot
# * on prend ensuite un texte quelconque qu'on découpe en phrase
# * pour chaque phrase qu'on rencontre, on remplace des mots au hasard par leur synonymes
# * chaque phrase sera présente une fois à l'identique et plusieurs fois avec des synonymes différents
# 
# L'objectif est alors de proposer une méthode pour reconstruire la base de synonymes.

# In[1]:


from jyquickhelper import add_notebook_menu
add_notebook_menu()


# ## Génération des données

# ### Construction de la base de synonymes

# In[2]:


from actuariat_python.data import wolf_xml
wolf_xml()


# In[3]:


import os
if not os.path.exists("wolf-1.0b4.xml"):
    raise FileNotFoundError("wolf-1.0b4.xml")
if os.stat("wolf-1.0b4.xml").st_size < 3000000:
    raise FileNotFoundError("Size of 'wolf-1.0b4.xml' is very small: {0}".format(os.stat("wolf-1.0b4.xml").st_size))


# In[4]:


from actuariat_python.data import enumerate_wolf_synonyms
for syn in enumerate_wolf_synonyms("wolf-1.0b4.xml", errors="ignore"):
    print(syn)
    break


# On passe en revue toute la basse (il y a environ 120.000 lignes) et on s'arrête après 10000 synonymes car sinon, cela prend un temps fou.

# In[5]:


allsyn = {}
for line, syn in enumerate(enumerate_wolf_synonyms("wolf-1.0b4.xml")):
    if line % 10000 == 0: print("line", line, "allsyn", len(allsyn))
    clean = [_.lower() for _ in syn if " " not in _]
    if len(clean) > 1:
        for word in clean:
            if word not in allsyn:
                allsyn[word] = set(clean)
                continue
            else:
                for cl in clean:
                    allsyn[word].add(cl)
    if len(allsyn) > 10000:
        break
len(allsyn)


# On affiche les premiers groupes :

# In[6]:


i = 0
for k, v in allsyn.items():
    print(k,v)
    i += 1
    if i > 10:
        break


# ### Génération d'une base de phrases modifiées
# 
# On utilise [Zadig](https://fr.wikipedia.org/wiki/Zadig).

# In[7]:


import urllib.request
with urllib.request.urlopen("http://www.gutenberg.org/cache/epub/4647/pg4647.txt") as u:
    content = u.read()


# In[8]:


char = content.decode(encoding="utf-8")


# On découpe en mot.

# In[9]:


import re
reg = re.compile("([- a-zA-Zâàäéèëêîïôöùûü']+)")
phrases = [_.lower() for _ in reg.findall(char)]


# In[10]:


for i, phrase in enumerate(phrases):
    if i >= 990:
        print(phrase)
    if i >= 1000:
        break


# On génère les phrases modifiées :

# In[11]:


import random
def modification(phrase, allsyn, nmax=10):
    mots = phrase.split()
    options = []
    nb = 1
    for mot in mots:
        if mot in allsyn:
            options.append(list(set([mot] + list(allsyn[mot]))))
        else:
            options.append([mot])
        nb *= len(options[-1])
    if nb == 1:
        return [phrase]
    else:
        res = []
        for i in range(0, min(nmax, nb//2+1, nb)):
            sol = []
            for mot in options:
                h = random.randint(0, len(mot)-1)
                sol.append(mot[h])
            res.append(sol)
        return res
            
modification("chatouiller le cérébral", allsyn)


# On traite tous les mots :

# In[12]:


len(phrases)


# In[13]:


with open("zadig_augmente.txt", "w", encoding="utf-8") as f:
    total = 0
    init = 0
    for i, phrase in enumerate(phrases):
        augm = modification(phrase, allsyn)
        init += 1
        for au in augm:
            f.write(" ".join(au) + "\n")
            total += 1
"total", total, "initial", init


# ## Exercice : retrouver une partie des synonymes à partir du dernier fichier créé
# 
# Le fichier utilisé peut être généré à partir du code précédent ou utiliser cette version : [zadig_augmente.zip](http://www.xavierdupre.fr/enseignement/complements/zadig_augmente.zip).

# In[14]:


from pyensae.datasource import download_data
download_data("zadig_augmente.zip")


# On lit le texte pour voir un peu de qu'il contient :

# In[15]:


with open("zadig_augmente.txt", "r", encoding="utf-8") as f:
    lines = f.readlines()
len(lines)


# In[16]:


lines[10000:10005]


# On supprime les caractères indésirables avec la fonction [strip](https://docs.python.org/3.4/library/stdtypes.html#str.strip) :

# In[17]:


lines = [ line.strip("\n \t\r") for line in lines ]
lines[10000:10005]


# Pour détecter les synonymes, on imagine le processus suivant qu'on pourra améliorer au vu des premiers résultats :

# In[18]:


from pyquickhelper.helpgen import NbImage
NbImage("synoym.png")


# On supprime les phrases qui incluent des espaces entre chaque lettre. On a la choix entre les garder, les jeter ou les corriger. On choisit l'option la plus simple : les jeter quitte à y revenir plus tard. Si un espace existe en chaque lettre, cela veut que le nombre de lettres (sans espace) est égal au nombre de mot.

# In[19]:


garde = []
for line in lines:
    nbchar = len(line.replace(" ", ""))
    nbword = len(line.split())
    if nbchar > nbword:
        garde.append(line)
len(garde), len(lines)


# In[20]:


garde[:5]


# On découpe en mot avec la fonction [split](https://docs.python.org/3.4/library/stdtypes.html#str.split) :

# In[21]:


mots = [line.split() for line in garde]
mots[:5]


# On code les petits bouts :

# In[22]:


def pourcentage_mots_communs(p1, p2):
    n = 0
    for p in p1:
        if p in p2:
            n += 1
    return n * 1.0 / max(len(p1),len(p2))

p1, p2 = mots[0:2]
print(p1)
print(p2)
pourcentage_mots_communs(p1, p2)


# Et l'autre :

# In[23]:


def mots_differents(p1, p2):
    diff1 = []
    for p in p1:
        if p not in p2:
            diff1.append(p)
    diff2 = []
    for p in p2:
        if p not in p1:
            diff2.append(p)
    return diff1, diff2

#p1, p2 = mots[0:2]
p1, p2 = ["a", "a", "b", "c"], ["a", "f", "g"]
print(p1)
print(p2)
mots_differents(p1, p2)


# Pour la boucle final, on procède par étape, on affiche des résultats intermédiaires : 

# In[24]:


for i in range(0,len(mots)):
    p1 = mots[i]
    p2 = mots[i+1]
    print("----")
    print(p1)
    print(p2)
    print(pourcentage_mots_communs(p1,p2))
    print(mots_differents(p1,p2))
    if i > 3 :
        break


# On complète cette première boucle en supposant qu'une bonne paire de synonymes doit provenir d'une paire de mots ayant en pourcentage de mots communs supérieur à 50% et une seule paire de synonymes :

# In[25]:


compteur = {}
for i in range(0,len(mots)-1):
    p1 = mots[i]
    p2 = mots[i+1]
    if pourcentage_mots_communs(p1, p2) >= 0.5:
        diff1, diff2 = mots_differents(p1, p2)
        if len(diff1) == len(diff2) == 1:
            m1 = diff1[0]
            m2 = diff2[0]
            l = [m1, m2]
            l.sort()  # on trie les deux mots pour aggréger les deux paires m1, m2 et m2, m1
            key = tuple(l)  # une liste ne peut être la clé d'un dictionnaire
            if key not in compteur:
                compteur[key] = 0
            compteur[key] += 1


# On regarde le résultat en triant par ordre de fréquence décroissante :

# In[26]:


l = [ (compteur[k],k) for k in compteur]
l[:5]


# In[27]:


l.sort(reverse=True)
l[:20]


# ## Vérifications
# 
# Ces mots ne paraissent pas très pertinents au regard de la langue française. La méthode est néanmoins valable si elle permet de retrouver des synonymes qui ont permis de générer ce texte *zadig_augmente.txt*.

# In[28]:


allsyn.get("lombric", None)


# In[29]:


allsyn.get("flétan", None)


# In[30]:


allsyn.get("niet", None)


# ## Améliorations

# ### stop words

# In[31]:


stopwords = {"de", "ou", "a", "un", "du"}
compteur = {}
for i in range(0,len(mots)-1):
    p1 = mots[i]
    p2 = mots[i+1]
    if pourcentage_mots_communs(p1, p2) >= 0.5:
        diff1, diff2 = mots_differents(p1, p2)
        if len(diff1) == len(diff2) == 1:
            m1 = diff1[0]
            m2 = diff2[0]
            if m1 not in  stopwords and m2 not in stopwords:
                l = [m1, m2]
                l.sort()
                key = tuple(l)
                if key not in compteur:
                    compteur[key] = 0
                compteur[key] += 1


# In[32]:


l = [ (compteur[k],k) for k in compteur]
l.sort(reverse=True)
l[:20]


# ### prise en compte de plus ligne
# 
# Et pas seulement deux lignes consécutives

# In[33]:


stopwords = {"de", "ou", "a", "un", "du"}
compteur = {}
for i in range(0,len(mots)-1):
    p1 = mots[i]
    for k in range(0,10):
        if i+k >= len(mots):
            break
        p2 = mots[i+k]
        if pourcentage_mots_communs(p1, p2) >= 0.5:
            diff1, diff2 = mots_differents(p1, p2)
            if len(diff1) == len(diff2) == 1:
                m1 = diff1[0]
                m2 = diff2[0]
                if m1 not in  stopwords and m2 not in stopwords:
                    l = [m1, m2]
                    l.sort()
                    key = tuple(l)
                    if key not in compteur:
                        compteur[key] = 0
                    compteur[key] += 1


# In[34]:


l = [ (compteur[k],k) for k in compteur]
l.sort(reverse=True)
l[:20]


# Ca ne change pas grand compte tenu de la façon dont la base a été construite.

# In[35]: